-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.type = "auxilliary"
M.relevantDevice = nil;

M.throttle_FL = 0
M.throttle_FR = 0
M.throttle_RL = 0
M.throttle_RR = 0

M.Debug = {}
M.Debug.FL = 0
M.Debug.FR = 0
M.Debug.RL = 0
M.Debug.RR = 0

--shorter functions for increased performance
local abs = math.abs
local min = math.min
local max = math.max
local sqrt = math.sqrt
local acos = math.acos
local clamp = clamp
local pi = math.pi
local pow = math.pow



-- debugging functions
local function length(array)
	local c = 0
	for _ in pairs(array) do c = c + 1 end
	return c
end


local function average(values)
	
	local sum = 0
	
	for k,v in pairs(values) do
		sum = sum + v
	end
	
	return sum / length(values)
	
end


-- this is an object to pass debug data to the UI via the electrics stream
local saberData = {}

-- caching these because it's easier to read
-- also a stored reference has got to be faster than a string search
local motor_FL = powertrain.getDevice("frontLeftMotor")
local motor_FR = powertrain.getDevice("frontRightMotor")
local motor_RL = powertrain.getDevice("rearLeftMotor")
local motor_RR = powertrain.getDevice("rearRightMotor") 

local wheel_RL = nil
local wheel_RR = nil
local wheel_FL = nil
local wheel_FR = nil

-- pid steering will be disabled if all these cutoffs are satisfied
local cutoffMinSpeed = 10/3.6 -- m/s, not km/h
local cutoffMinGForce = 0.2
local cutoffYawRate = 0.1


local wheelBase = 0
local invWheelBase = 0

-- local yawSmooth = nil
local steerSmooth = nil
local speed = 0
local desiredYawRateSteering = 0
local xGForce = 0
 

local configData = nil

local escPulse = 0
local tcsPulse = 0

--variables for ESC functionality
local wheelCache = {} --cache for holding all our wheels
local wheelIdCache = {}
local wheelCacheSize = 0
local wheelNameCache = {}
local otherWheelOnAxle = {}
local crossWheels = {}
local tcsWheelBrakeTorques = {}
local revLimiterEngines = {}
local revLimiterEngineCount = 0

local yawSmooth = nil --exponential smoothing for the yaw rate
local desiredYawSmooth = nil --exponential smoothing for the yaw rate
local invSquaredCharacteristicSpeed = 0 --pre calculated, used for desired yaw calculation

local frontLeftWheelId = nil
local frontRightWheelId = nil
local rearRightWheelId = nil
local rearLeftWheelId = nil

local escConfigurations = {}
local currentESCConfiguration = nil
local currentESCConfigurationKey = 1
local lastESCConfigurationKey = -1

local wheelBase = 0 --m
local invWheelBase = 0
local distanceCOGFrontAxle = 0 --m
local distanceCOGRearAxle = 0 --m
local trackWidth = 0
local trackWidthRefLeft = 0
local trackWidthRefRight = 0
local mass = 0 --kg

local initialWheelCount = 0
local escFailure = false

local desiredYawRateSteering = 0
local desiredYawRateAcceleration = 0
local yawRate = 0
local speed = 0 --m/s
local wheelAngleFront = 0
local wheelAngleRear = 0
local desiredYawRate = 0
local yawDifference = 0
local escEnableThreshold = 6 --m/s

local lateralAccSmoother = nil
local tempRevLimiterTimer = 0
local tempRevLimiterActive = false
local accelerationSensorSpeed = 0
local throttleFactorWheelLookup = {}
local throttleFactorData = {}

local tcsDeactivateThreshold = 0.01
local tcsDeactivateSpeedThreshold = 0.55

local escActive = false
local tcsActive = false
local warningLightsDelayTime = 0.15 --s
local warningLightsTimer = 0 --s
local offColor = nil

--ESC calibbration
local escMeasuringStepThreshold = 3500
local escSettlingStepThreshold = 1500
local escMeasuringStepCounter = 0
local escSettlingStepCounter = 0
local stiffnessFrontSum = 0
local stiffnessRearSum = 0

--toggle debug mode
local isDebugMode = 0
local calibrateESC = nop
local hasRegisteredQuickAccess = false

local wheelSlips = {}
local wheelBrakeFactors = {}
local escInfoData = {}


M.debugData = {
  esc = {
    steeringAngle = 0,
    yawRate = 0,
    desiredYawRate = 0,
    difference = 0,
    desiredYawRateAcceleration = 0,
    desiredYawRateSteering = 0,
    speed = 0
  },
  tcs = {
    throttleFactor = 0,
    wheelBrakeFactors = wheelBrakeFactors,
    allWheelSlip = 0,
    wheelSlips = wheelSlips,
    slipThreshold = 0
  }
}


--  variables for Saber specific settings

local crossAxleMaxFactor = 1 -- how much influence the cross-axle vectoring has at full throttle
local crossAxleMinFactor = 0.05 -- how much influence the cross-axle vectoring has at zero throttle
local interAxleSteerFactor = 1 -- how much influence the inter-axle vectoring has
local interAxleGForcePower = 1/2 -- how the influence should be shaped by g-force
local interAxleCriticalGPoint = 10 -- where the 1,1 point lies on the power curve
local interAxleCriticalGPointInverse = 0

local throttleShapingPower = 1

local maxThrottleDivergence = 0.5 -- how much the throttle output values can depart from the throttle input

local skidSteerStationaryYawRate = 10 -- how fast the vehicle can rotate in skid-steer mode at a standstill
local skidSteerBoost = 1 -- how much to add to the fast yaw rate in skid-steer mode


local throttleDirection = 0


-- yer basic PID settings
local yaw_PID_gain = 1 -- strength
local yaw_P = 6 -- aggression
local yaw_I = 1 -- error correction
local yaw_I_max = 10 -- max error correction
local yaw_D = 0.01 -- damping

-- yer basic PID object
local yaw_PID = {G = yaw_PID_gain, P = yaw_P, I = yaw_I, D = yaw_D, SP = 0, MPV = 0, MPV_prev = 0, I_sum = 0}


-- yer basic PID loop
local function update_PID(pid, sp, mpv, dt)
	
	pid.MPV_prev = pid.MPV
	pid.MPV = mpv
	pid.SP = sp
	
	local D_rate = (pid.MPV - pid.MPV_prev)/dt
	local P_error = pid.SP - pid.MPV
	pid.I_sum = pid.I_sum + P_error * dt
	
	return pid.G * (pid.P * P_error + pid.I * pid.I_sum - pid.D * D_rate)
	
end

-- this is to reset the PID integral whenever we change between forward, reverse and neutral
-- this is set manually and not from electrics.values.gear so that it can be overridden manually
local function setThrottleDirection(dir)
	
	if throttleDirection ~= dir then
		yaw_PID.I_sum = 0
		throttleDirection = dir
	end
	
end


local function cutoffCheck(speed, gforce, yaw, throttle, brake, parking)
	
		if speed < cutoffMinSpeed and gforce < cutoffMinGForce and yaw < cutoffYawRate and throttle == 0 and parking == 0 then
			return true
			
		elseif speed < cutoffMinSpeed and brake == 1 and parking == 0 then
			return true
		
		end
		
		return false
end


local function applyMotorThrottle(t, motor)
		
		-- reverse motor if throttle value is negative
		motor.motorDirection = fsign(t)
		-- set throttle to positive
		t = abs(t)
		
		-- return throttle and enjoy
		return t
	
end


local function normalisePower(values, p, throttle)
	
	local maxT = -1
	local minT = 1
	
	local tHeadroom = min(1 - throttle, maxThrottleDivergence)
	local tFootroom = min(1 + throttle, maxThrottleDivergence)
	
	for _,v in pairs(values) do
		maxT = max(v, maxT)
		minT = min(v, minT)
	end
	
	
	for k,_ in pairs(values) do
	
	
		if values[k] > throttle then
			
			if maxT > throttle + tHeadroom then
				values[k] = throttle + (values[k] - throttle) / (maxT - throttle) * tHeadroom
				
			end
			
			
		elseif values[k] < throttle then
			
			if minT < throttle - tFootroom then
				values[k] = throttle - (throttle - values[k]) / (throttle - minT) * tFootroom
				
			end
			
		end
		
		if values[k] > throttle then
			values[k] = throttle + pow(abs(values[k] - throttle), p) * tHeadroom
			
		elseif values[k] < throttle then
			values[k] = throttle - pow(abs(throttle - values[k]), p) * tFootroom
			
		end
			
			
		
	end
	
	
	-- local adjustSum = 0
	
	-- for _,v in pairs(values) do
		-- adjustSum = adjustSum + v - throttle
	-- end
	
	-- print(maxT)
	-- print(values.RR)
	
end



local function interAxleSteering(values, throttle, steer)
	
	
	
	for k,v in pairs(values) do
		
		if ((k == "FL" or k == "FR") and steer > 0) or ((k == "RL" or k == "RR") and steer < 0) then
			values[k] = v * (1 + abs(steer) * (1 - abs(v)))
		else
			values[k] = v / (1 + abs(steer))
			
		end
	end
	
end



local function updateGFX(dt)
	warningLightsTimer = warningLightsTimer + dt

	if warningLightsTimer >= warningLightsDelayTime then
		escPulse = escActive and bit.bxor(escPulse, 1) or 0
		tcsPulse = tcsActive and bit.bxor(tcsPulse, 1) or 0
		warningLightsTimer = 0
	end

	if currentESCConfiguration.overrideESCPulse then
		escPulse = currentESCConfiguration.overrideESCPulse
	end

	if currentESCConfiguration.overrideTCSPulse then
		tcsPulse = currentESCConfiguration.overrideTCSPulse
	end

	if escFailure then
		escPulse = 1
		tcsPulse = 1
	end

	electrics.values.esc = escPulse
	electrics.values.tcs = tcsPulse
	electrics.values.escActive = escActive
	electrics.values.tcsActive = tcsActive

	M.debugData.esc.yawRate = yawRate
	M.debugData.esc.desiredYawRate = desiredYawRate
	M.debugData.esc.difference = yawDifference

	table.clear(wheelSlips)
	table.clear(wheelBrakeFactors)

	for i = 1, wheelCacheSize, 1 do
		local wheel = wheelCache[wheelIdCache[i]]
		wheelSlips[wheel.name] = wheel.tractionControlLastSlip
		wheelBrakeFactors[wheel.name] = wheel.tractionControlBrakeFactor
	end

	local throttleFactors = {}
	for k, v in pairs(throttleFactorData) do
		throttleFactors[k] = v.throttleFactor
	end

	M.debugData.tcs.throttleFactors = throttleFactors
	M.debugData.tcs.wheelBrakeFactors = wheelBrakeFactors
	M.debugData.tcs.wheelSlips = wheelSlips
	M.debugData.tcs.slipThreshold = currentESCConfiguration.slipThreshold

	if streams.willSend("escInfo") then
		escInfoData.ledColor = (escPulse > 0 or tcsPulse > 0) and offColor or currentESCConfiguration.activeColor
		gui.send("escInfo", escInfoData)
	end
end



local function updateWheelsIntermediate(dt)
	
	if electrics.values.gear == "N" or electrics.values.gear == "P" then
		setThrottleDirection(0)
		
		electrics.values.throttle_FL = 0
		electrics.values.throttle_FR = 0
		electrics.values.throttle_RL = 0
		electrics.values.throttle_RR = 0
	end
	
	
	local throttleInput = electrics.values.throttle_input
	local brakeInput = electrics.values.brake_input
	local parkingInput = electrics.values.parkingbrake_input
	
	-- getting these smoothed values here so there's no discontinuity in data collection for the smoothers
	local yawRate = yawSmooth:get(obj:getYawAngularVelocity())
	local steeringInput = steerSmooth:get(input.steering)
	
	if electrics.values.gear == "D" or electrics.values.gear == "R" then
		
		if electrics.values.gear == "R" then
			setThrottleDirection(-1)
			
		elseif electrics.values.gear == "D" then
			throttleInput = throttleInput - electrics.values.clutch_input
			
			if throttleInput > 0 then
				setThrottleDirection(1)
				
			elseif throttleInput < 0 then
				setThrottleDirection(-1)
				throttleInput = abs(throttleInput)
				
			end
			
		end
		
		
		-- desired yawrate will be based on wheelbase and speed
		local speed = electrics.values.wheelspeed
		local wheelFront = steeringInput < 0 and wheel_FL or wheel_FR
		local wheelRear = steeringInput < 0 and wheel_RL or wheel_RR
		
		
		-- stock ESC steering magic
		local wheelAngleFront = acos(obj:nodeVecPlanarCosRightForward(wheelFront.node1, wheelFront.node2))
		local wheelAngleRear = acos(obj:nodeVecPlanarCosRightForward(wheelRear.node1, wheelRear.node2))
		if wheelAngleFront > 1.5708 then
			wheelAngleFront = (pi - wheelAngleFront)
		end
		wheelAngleFront = wheelAngleFront * fsign(-steeringInput)
		if wheelAngleFront ~= wheelAngleFront then
			wheelAngleFront = 0
		end

		if wheelAngleRear > 1.5708 then
			wheelAngleRear = (pi - wheelAngleRear)
		end
		wheelAngleRear = wheelAngleRear * fsign(-steeringInput)
		if wheelAngleRear ~= wheelAngleRear then
			wheelAngleRear = 0
		end

		M.wheelAngleFront = wheelAngleFront
		M.wheelAngleRear = wheelAngleRear
		
		
		-- calculate expected yaw rate based on steering angle
		-- need to invert this when reversing because the PID doesn't like reversing its output
		desiredYawRateSteering = -((wheelAngleFront * invWheelBase) * (speed/(1 + (speed * speed * invSquaredCharacteristicSpeed)))) * throttleDirection
		
		
		-- parking brake maps to skid-steer mode when in drive
		if parkingInput > 0 then
			-- we don't want to use the actual parking brake and this seems to intercept it before it reaches the actual wheels and I'm not asking questions at this point
			input.parkingbrake = 0
			
			-- pick the stationary yaw rate, or the boosted fast yaw rate, whichever is higher
			desiredYawRateSteering = max(abs(skidSteerStationaryYawRate * parkingInput * steeringInput), abs(desiredYawRateSteering * (1 + skidSteerBoost * parkingInput))) * fsign(desiredYawRateSteering)
			
		end
		
		
		-- adjust the PIDSteer based on speed
		-- to get rid of juddering at lower speeds
		-- and improve response at high speeds
		local PIDSpeedFactor = (speed / currentESCConfiguration.characteristicSpeed)
		
		-- need to invert it again at this point because we lied to the PID, so it gives us an inverted reponse back
		local PIDOutput = update_PID(yaw_PID, desiredYawRateSteering, yawRate, dt)
		local PIDSteer = PIDOutput * max(PIDSpeedFactor, fsign(parkingInput)) * throttleDirection
		
		-- lerp between factors depending on throttle/parking brake, prevents over-intervention at low throttle inputs
		local crossAxleFactor = (max(throttleInput, parkingInput) * crossAxleMaxFactor + (1 - throttleInput) * crossAxleMinFactor)
		 
		-- reduce cross-axle vectoring at higher speeds
		local crossAxleDelta = PIDSteer * crossAxleFactor
		
		-- inter axle scaling is based entirely on lateral g-force and the desired steering results
		-- no need to look at steering angle or understeer/oversteer which are misleading names to say the least
		-- this is a scale because just putting a negative or positive throttle won't necessarily
		-- improve traction, we need to reduce or boost the power overall regardless of direction
		local xGForce = sensors.gx2
		
		
		if cutoffCheck(speed, xGForce, yawRate, throttleInput, brakeInput, parkingInput) then
			
			yaw_PID.I_sum = 0
			return
			
		end
		
		
		
		local interAxleSteerGFactor = pow(interAxleCriticalGPointInverse * abs(xGForce), interAxleGForcePower)
		local interAxleSteer = PIDSteer * interAxleSteerFactor * interAxleSteerGFactor
		
		--crossAxleDelta = crossAxleDelta-- * interAxleSteerGFactor
		
		-- negating this number switches from oversteer to understeer
		if fsign(xGForce) * throttleDirection < 0 then
			interAxleSteer = -interAxleSteer			
		end
		
		
		-- apply steering factors and reversing
		
		local throttleResults = {}
		throttleResults.FL = (throttleInput + crossAxleDelta) * throttleDirection
		throttleResults.FR = (throttleInput - crossAxleDelta) * throttleDirection
		throttleResults.RL = (throttleInput + crossAxleDelta) * throttleDirection
		throttleResults.RR = (throttleInput - crossAxleDelta) * throttleDirection
		
		
		normalisePower(throttleResults, throttleShapingPower, throttleInput * throttleDirection)
		
		
		interAxleSteering(throttleResults, throttleInput, interAxleSteer)
		
		--print(average(throttleResults))
		
		-- M.Debug.FL = throttleResults.FL
		-- M.Debug.FR = throttleResults.FR
		-- M.Debug.RL = throttleResults.RL
		-- M.Debug.RR = throttleResults.RR

		electrics.values.saberData = {}
		
		electrics.values.saberData.wat = "WAT"
		
		electrics.values.saberData.FL = throttleResults.FL
		electrics.values.saberData.FR = throttleResults.FR
		electrics.values.saberData.RL = throttleResults.RL
		electrics.values.saberData.RR = throttleResults.RR
		
		
		
		
		-- apply all throttle values
		electrics.values.throttle_FL = applyMotorThrottle(throttleResults.FL, motor_FL)
		electrics.values.throttle_FR = applyMotorThrottle(throttleResults.FR, motor_FR)
		electrics.values.throttle_RL = applyMotorThrottle(throttleResults.RL, motor_RL)		
		electrics.values.throttle_RR = applyMotorThrottle(throttleResults.RR, motor_RR)
		
		
	end
	
	
	
	
	
end


-------------------------------------
------- Stock Boilerplate -----------
-------------------------------------


local function doCalibration()
	if M.doMeasure then
		escMeasuringStepCounter = escMeasuringStepCounter + 1

		local velocity = vec3(obj:getVelocity())
		local directionVector = vec3(obj:getDirectionVector())
		local actualVelocity = (directionVector:dot(velocity) / (directionVector:length() * directionVector:length()) * directionVector):length()
		local velocityVector = vec3(velocity.x, velocity.y, 0)
		local dot = velocityVector:dot(directionVector)
		local floatAngle = acos(dot / (directionVector:length() * velocityVector:length()))

		local yawRateCalibration = abs(yawRate) * -1
		local wheelAngleFrontCalibration = abs(wheelAngleFront) * -1
		local wheelAngleRearCalibration = abs(wheelAngleRear) * -1

		local stiffnessFront = (distanceCOGRearAxle * mass * yawRateCalibration * actualVelocity) / ((distanceCOGFrontAxle + distanceCOGRearAxle) * (wheelAngleFrontCalibration - (distanceCOGFrontAxle * yawRateCalibration / actualVelocity) - floatAngle))
		local stiffnessRear = ((yawRateCalibration * mass * actualVelocity) - stiffnessFront * (wheelAngleFrontCalibration - (distanceCOGFrontAxle * yawRateCalibration / actualVelocity) - floatAngle)) / (wheelAngleRearCalibration + (distanceCOGRearAxle * yawRateCalibration / actualVelocity) - floatAngle)

		stiffnessFrontSum = stiffnessFrontSum + stiffnessFront
		stiffnessRearSum = stiffnessRearSum + stiffnessRear

		if escMeasuringStepCounter >= escMeasuringStepThreshold then
			M.stiffnessFront = stiffnessFrontSum / escMeasuringStepCounter
			M.stiffnessRear = stiffnessRearSum / escMeasuringStepCounter
			stiffnessFrontSum = 0
			stiffnessRearSum = 0
			M.doMeasure = false
			M.calibrationMeasurementReady = true
			escMeasuringStepCounter = 0
		end
	end

	if M.doSettle then
		escSettlingStepCounter = escSettlingStepCounter + 1
		if escSettlingStepCounter >= escSettlingStepThreshold then
			M.doSettle = false
			M.calibrationSettled = true
			escSettlingStepCounter = 0
		end
	end
end



local function startESCCalibration()
	M.calibrationMeasurementReady = false
	M.doMeasure = false
	M.doSettle = false
	M.calibrationSettled = false
	stiffnessFrontSum = 0
	stiffnessRearSum = 0
	M.stiffnessFront = 0
	M.stiffnessRear = 0
	escMeasuringStepCounter = 0
	escSettlingStepCounter = 0
	calibrateESC = doCalibration
end



local function stopESCCalibration()
	calibrateESC = nop
end



local function getCarData()
	if mass <= 0 then
		return nil
	end

	return {
		wheels = wheelCache,
		frontLeftWheelId = frontLeftWheelId,
		frontRightWheelId = frontRightWheelId,
		rearLeftWheelId = rearLeftWheelId,
		rearRightWheelId = rearRightWheelId,
		wheelBase = wheelBase,
		trackWidth = trackWidth,
		distanceCOGRearAxle = distanceCOGRearAxle,
		distanceCOGFrontAxle = distanceCOGFrontAxle,
		mass = mass
	}
end




local function getCurrentConfigData()
	return currentESCConfiguration
end



local function sanitizeConfiguration(config, name)
	config.name = name
	config.escEnabled = config.escEnabled or false
	config.activeColor = config.activeColor or "98FB00"
	config.characteristicSpeed = config.characteristicSpeed or 0

	config.proportionalFactor = config.proportionalFactor or 1
	config.integralFactor = config.integralFactor or 0
	config.maxIntegralPart = config.maxIntegralPart or 2

	config.maxSideAcceleration = config.maxSideAcceleration or 0
	config.brakeForceMultiplier = config.brakeForceMultiplier or 0
	config.escThreshold = config.escThreshold or 0

	config.skewStiffnessFront = config.skewStiffnessFront or 1
	config.skewStiffnessRear = config.skewStiffnessRear or 1
	config.desiredYawRateSmoothing = config.desiredYawRateSmoothing or 500

	config.slipThreshold = config.slipThreshold or 0
	config.tcsWheelSpeedThreshold = config.tcsWheelSpeedThreshold or 10

	config.throttleProportionalFactor = config.throttleProportionalFactor or 1.5
	config.throttleIntegralFactor = config.throttleIntegralFactor or 1
	config.throttleIntegralReductionFactor = config.throttleIntegralReductionFactor or config.slipThreshold * 2

	config.brakingProportionalFactor = config.brakingProportionalFactor or 1.2
	config.brakingIntegralFactor = config.brakingIntegralFactor or 0

	config.maxBrakingFactor = clamp(config.maxBrakingFactor or 0, 1, 0)
	config.minThrottleFactor = clamp(config.minThrottleFactor or 1, 1, 0)

	config.brakeThrottleSwitchThreshold = config.brakeThrottleSwitchThreshold or 10

	return config
end



local function preCalculate()
	invSquaredCharacteristicSpeed = 1 / (currentESCConfiguration.characteristicSpeed * currentESCConfiguration.characteristicSpeed)
end



local function setESCMode(key)
	if currentESCConfiguration == nil then --if we don't have any current config, the vehicle does not have esc at all, abort here
		return
	end

	if initialWheelCount ~= wheels.wheelCount then --if we have detached wheels, the esc is disabled anyway, no need to switch modes anymore
		return
	end

	currentESCConfigurationKey = key
	if currentESCConfigurationKey > #escConfigurations then
		currentESCConfigurationKey = 1
	end

	currentESCConfiguration = escConfigurations[currentESCConfigurationKey] --load new esc config

	preCalculate() -- make sure to update our precalculated values with the new config

	lastESCConfigurationKey = currentESCConfigurationKey
	gui.message(currentESCConfiguration.name, 5, "vehicle.esc.mode")
end



local function toggleESCMode()
	--sounds.playSoundOnceAtNode("event:>fabian_test", 0, 1)
	local key = currentESCConfigurationKey + 1
	setESCMode(key)
end



local function nextESCMode()
	local key = min(currentESCConfigurationKey + 1, #escConfigurations)
	setESCMode(key)
end



local function previousESCMode()
	local key = max(currentESCConfigurationKey - 1, 1)
	setESCMode(key)
end



local function registerQuickAccess()
	if not hasRegisteredQuickAccess then
		core_quickAccess.addEntry(
			{
				level = "/",
				generator = function(entries)
					table.insert(entries, {title = "ui.radialmenu2.ESC", priority = 40, ["goto"] = "/esc/", icon = "radial_regular_esc"})
				end
			}
		)

		core_quickAccess.addEntry(
			{
				level = "/esc/",
				generator = function(entries)
					for k, v in pairs(escConfigurations) do
						local entry = {
							title = "ui.radialmenu2.ESC." .. v.name:gsub(" ", "_"),
							icon = "radial_" .. string.lower(v.name:gsub(" ", "_")),
							onSelect = function()
								controller.getController("esc").setESCMode(k)
								return {"reload"}
							end
						}
						if currentESCConfiguration == v then
							entry.color = "#ff6600"
						end
						table.insert(entries, entry)
					end
				end
			}
		)
		hasRegisteredQuickAccess = true
	end
end



local function calculateCharacteristicSpeed(config)
	local eg = (mass * (config.skewStiffnessRear * distanceCOGRearAxle - config.skewStiffnessFront * distanceCOGFrontAxle)) / (config.skewStiffnessFront * config.skewStiffnessRear * wheelBase)
	local characteristicSpeed = sqrt(wheelBase / abs(eg + 1e-30)) --guard against infinity
	if isDebugMode and config.escEnabled then
		log("D", "ESC", string.format("Calculated EG: %s --> %.6f", config.name, eg))
		log("D", "ESC", string.format("Calculated characteristic speed: %s --> %.2f m/s", config.name, characteristicSpeed))
		if eg < 0 then
			log("W", "ESC", string.format("Calculated EG (%s) is lower than 0 (oversteery car setup), ESC might not work perfectly!", config.name))
		end
	end
	return characteristicSpeed
end



local function calculateAxleDistances()
	wheelBase = obj:nodeLength(wheelCache[frontRightWheelId].node1, wheelCache[rearRightWheelId].node1) --calculate wheelbase from the distance of the front and rear wheels
	invWheelBase = 1 / wheelBase

	local tmp = vec3(0, 0, 0)
	local totalMass = 0
	for _, v in pairs(v.data.nodes) do
		tmp = tmp + vec3(v.pos) * v.nodeWeight
		totalMass = totalMass + v.nodeWeight
	end

	local realCOG = tmp / totalMass

	--Find the positions of the front and rear axle
	local frontAxlePos = 0
	local rearAxlePos = 0
	local twLeft = 0
	local twRight = 0
	for _, n in pairs(v.data.nodes) do
		if n.cid == wheelCache[frontRightWheelId].node1 then
			frontAxlePos = n.pos.y
			twRight = n.pos.x
		elseif n.cid == wheelCache[rearRightWheelId].node1 then
			rearAxlePos = n.pos.y
		elseif n.cid == wheelCache[frontLeftWheelId].node1 then
			twLeft = n.pos.x
		end
	end

	distanceCOGFrontAxle = abs(realCOG.y - frontAxlePos)
	distanceCOGRearAxle = abs(realCOG.y - rearAxlePos)

	trackWidth = abs(twLeft - twRight)
	local refNodeX = v.data.nodes[v.data.refNodes[0].ref].pos.x
	trackWidthRefLeft = abs(twLeft - refNodeX)
	trackWidthRefRight = abs(twRight - refNodeX)

	if isDebugMode then
		log("D", "ESC", "Distance COG to Rearaxle: " .. distanceCOGRearAxle .. " m")
		log("D", "ESC", "Distance COG to Frontaxle: " .. distanceCOGFrontAxle .. " m")
		log("D", "ESC", "Wheelbase: " .. wheelBase .. " m")
		log("D", "ESC", "Track width: " .. trackWidth .. " m")
	end
end



local function init(jbeamData)
	configData = jbeamData
end



local function initSecondStage()
	
	-- grab all the wheels and cache them
	for _,wd in pairs(wheels.wheels) do
		if (wd.name ==  "FL") then
			wheel_FL = wd
		elseif (wd.name == "FR") then
			wheel_FR = wd
		elseif (wd.name == "RL") then
			wheel_RL = wd
		elseif (wd.name == "RR") then
			wheel_RR = wd
		end
	end
	
	yawSmooth = newExponentialSmoothing(50)
	
	steerSmooth = newExponentialSmoothing(10)
	
	interAxleCriticalGPointInverse = 1 / interAxleCriticalGPoint
	
	--mass = totalMass
	
	--characteristicSpeed = calculateCharacteristicSpeed()
	
	--invSquaredCharacteristicSpeed = 1 / (characteristicSpeed * characteristicSpeed)
	
	escPulse = 0
	tcsPulse = 0
	escFailure = false
	M.pauseESCAction = false
	M.wheelAngleFront = 0
	M.wheelAngleRear = 0
	M.stiffnessFront = 0
	M.stiffnessRear = 0

	calibrateESC = nop
	M.updateWheelsIntermediate = nil
	M.updateGFX = nil

	wheelCache = table.new(4, 0)
	wheelCacheSize = 0
	wheelNameCache = {}
	tcsWheelBrakeTorques = {}
	accelerationSensorSpeed = 0

	--cache all wheels for easy access

	local escConfigs = configData
	
	local actionedWheelLookup = {}
	for _, wheelName in ipairs(escConfigs.actionedWheels) do
		actionedWheelLookup[wheelName] = true
	end

	for id, wd in pairs(wheels.wheels) do
		if actionedWheelLookup[wd.name] then
			local wheelCacheEntry = wd
			wheelCacheEntry.speedSmoother = newExponentialSmoothing(100)
			wheelCacheEntry.stabilityControlBrakeIntegral = 0
			wheelCacheEntry.tractionControlBrakeIntegral = 0
			wheelCacheEntry.tractionControlThrottleIntegral = 0
			wheelCacheEntry.tractionControlBrakeFactor = 0
			wheelCacheEntry.tractionControlLastSlip = 0
			wheelCacheEntry.tractionControlSpeedCorrectionFactor = 1
			--table.insert(wheelCache, wheelCacheEntry)
			wheelCache[id] = wheelCacheEntry
			wheelNameCache[wd.name] = id
			wheelCacheSize = wheelCacheSize + 1
			tcsWheelBrakeTorques[id] = 0
		end
	end

	yawSmooth = newExponentialSmoothing(50) --windows of 50 frames

	--we need to find the average wheel position (basically the "center" of our vehicle)
	local avgWheelPos = vec3(0, 0, 0)
	local wheelCount = 0

	if v.userSettings and v.userSettings.escConfig then
		escConfigs = tableMergeRecursive(escConfigs, v.userSettings.escConfig)
	end

	for _, wheelName in ipairs(escConfigs.actionedWheels) do
		if wheelNameCache[wheelName] == nil then
			log("W", "ESC", "Could not find wheel: " .. wheelName .. " defined in escConfig")
			M.update = nop
			M.graphicsStep = nop
			return
		end
		local wheelNodePos = v.data.nodes[wheelCache[wheelNameCache[wheelName]].node1].pos --find the wheel position
		avgWheelPos = avgWheelPos + wheelNodePos --sum up all positions
		wheelCount = wheelCount + 1
	end

	avgWheelPos = avgWheelPos / wheelCount --make the average of all positions

	local vectorForward = vec3(v.data.nodes[v.data.refNodes[0].ref].pos) - vec3(v.data.nodes[v.data.refNodes[0].back].pos) -- vec3(obj:getDirectionVector()) --vector facing forward
	local vectorUp = vec3(v.data.nodes[v.data.refNodes[0].up].pos) - vec3(v.data.nodes[v.data.refNodes[0].ref].pos)

	local vectorRight = vectorForward:cross(vectorUp) --vector facing to the right

	offColor = escConfigs.offColor or "343434"

	--iterate over all wheels that should be included in the esc
	for _, wheelName in ipairs(escConfigs.actionedWheels) do
		local wheelNodePos = vec3(v.data.nodes[wheelCache[wheelNameCache[wheelName]].node1].pos) --find the wheel position
		local wheelVector = wheelNodePos - avgWheelPos --create a vector from our "center" to the wheel
		local dotForward = vectorForward:dot(wheelVector) --calculate dot product of said vector and forward vector
		local dotLeft = vectorRight:dot(wheelVector) --calculate dot product of said vector and left vector

		if dotForward >= 0 then
			if dotLeft >= 0 then
				frontRightWheelId = wheelNameCache[wheelName] --this case can only mean it's our front right wheel
			else
				frontLeftWheelId = wheelNameCache[wheelName] -- ...
			end
		else
			if dotLeft >= 0 then
				rearRightWheelId = wheelNameCache[wheelName] -- ...
			else
				rearLeftWheelId = wheelNameCache[wheelName] -- ...
			end
		end
	end

	local throttleFactorAssingment
	if escConfigs.throttleFactorAssingment then
		throttleFactorAssingment = tableFromHeaderTable(escConfigs.throttleFactorAssingment)
	else
		throttleFactorAssingment = {}
		for _, wheelName in ipairs(escConfigs.actionedWheels) do
			table.insert(throttleFactorAssingment, {wheelName = wheelName, throttleFactorName = "throttleFactor"})
		end
	end

	throttleFactorWheelLookup = {}
	throttleFactorData = {}
	for _, v in pairs(throttleFactorAssingment) do
		throttleFactorWheelLookup[v.wheelName] = v.throttleFactorName
		if not throttleFactorData[v.throttleFactorName] then
			throttleFactorData[v.throttleFactorName] = {
				peakSlip = 0,
				integral = 0,
				throttleFactor = 1,
				tcsActive = false,
				allWheelSlipSmoother = newTemporalSmoothing(1, 1000),
				allWheelSlip = false
			}
		end
	end

	initialWheelCount = wheels.wheelCount

	--create an easy way to access the "other" wheel on an axle
	otherWheelOnAxle[frontLeftWheelId] = frontRightWheelId
	otherWheelOnAxle[frontRightWheelId] = frontLeftWheelId
	otherWheelOnAxle[rearLeftWheelId] = rearRightWheelId
	otherWheelOnAxle[rearRightWheelId] = rearLeftWheelId

	--dump(otherWheelOnAxle)
	wheelIdCache = {}
	for k, _ in pairs(wheelCache) do
		table.insert(wheelIdCache, k)
	end

	local tmpConfigs = {}
	for name, config in pairs(escConfigs.configurations) do
		if type(config) == "table" and config.escConfigurationEnabled then
			table.insert(tmpConfigs, sanitizeConfiguration(shallowcopy(config), name)) --we need to create copies of all our configurations as we are going to use them for saving a few values as well and we want a fresh copy every time we reset the vehicle
		end
	end

	local counter = 1
	escConfigurations = {}
	table.sort(
		tmpConfigs,
		function(a, b)
			return b.order > a.order
		end
	)
	for _, config in pairs(tmpConfigs) do
		escConfigurations[counter] = config
		counter = counter + 1
	end

	if lastESCConfigurationKey ~= -1 then
		currentESCConfigurationKey = lastESCConfigurationKey
	else
		currentESCConfigurationKey = escConfigs.defaultConfig or 1
		lastESCConfigurationKey = currentESCConfigurationKey
	end
	currentESCConfiguration = escConfigurations[currentESCConfigurationKey] --load the default configuration
	gui.message(currentESCConfiguration.name, 5, "vehicle.esc.mode")

	isDebugMode = escConfigs.isDebugMode > 0

	if isDebugMode then
		log("D", "ESC", "ESC configuration data:")
		log("D", "ESC", dumps(configData))
		log("D", "ESC", "Using ESC configuration: " .. currentESCConfiguration.name)
		log("D", "ESC", "Front Left wheel: " .. frontLeftWheelId)
		log("D", "ESC", "Front Right wheel: " .. frontRightWheelId)
		log("D", "ESC", "Rear Left wheel: " .. rearLeftWheelId)
		log("D", "ESC", "Rear Right wheel: " .. rearRightWheelId)
	end

	calculateAxleDistances()
	local statsObj = obj:calcBeamStats()
	mass = statsObj.total_weight --simply the mass of the car

	for _, config in pairs(escConfigurations) do
		if type(config) == "table" and config.escConfigurationEnabled and config.characteristicSpeed <= 0 then --calculate char. speed if no override is provided
			config.characteristicSpeed = calculateCharacteristicSpeed(config)
		end
	end

	desiredYawSmooth = newExponentialSmoothing(currentESCConfiguration.desiredYawRateSmoothing)
	lateralAccSmoother = newTemporalSmoothing(100, 100)

	preCalculate()

	--TCS
	crossWheels[frontLeftWheelId] = rearRightWheelId
	crossWheels[rearRightWheelId] = frontLeftWheelId
	crossWheels[frontRightWheelId] = rearLeftWheelId
	crossWheels[rearLeftWheelId] = frontRightWheelId

	local engines = powertrain.getDevicesByCategory("engine") --get all devices with "engine" category
	local blacklistedEngines = {}
	if escConfigs.blacklistedEngines and type(escConfigs.blacklistedEngines) == "table" then
		for _, v in pairs(escConfigs.blacklistedEngines) do
			blacklistedEngines[v] = true
		end
	end
	revLimiterEngines = {}
	for _, v in pairs(engines) do
		--make sure the device supports temp rev limiters, otherwise we'll just run into issues
		if v.setTempRevLimiter and v.resetTempRevLimiter and not blacklistedEngines[v.name] then
			table.insert(revLimiterEngines, v)
		end
	end
	revLimiterEngineCount = #revLimiterEngines
	tempRevLimiterActive = false
	tempRevLimiterTimer = 0

	registerQuickAccess()
	
	M.updateWheelsIntermediate = updateWheelsIntermediate
	M.updateGFX = updateGFX
	
	
	-- PID settings
	yaw_PID.G = yaw_PID_gain
	yaw_PID.P = getCurrentConfigData().steerPID_P or 0
	yaw_PID.I = getCurrentConfigData().steerPID_I or 0
	yaw_PID.I_max = yaw_I_max
	yaw_PID.D = getCurrentConfigData().steerPID_D or 0
	
	-- vector settings
	crossAxleMaxFactor = getCurrentConfigData().steerXAMax
	crossAxleMinFactor = getCurrentConfigData().steerXAMin
	interAxleSteerFactor = getCurrentConfigData().steerIAGain
	interAxleCriticalGPoint = getCurrentConfigData().steerIACriticalG
	maxThrottleDivergence = getCurrentConfigData().maxDivergence
	skidSteerStationaryYawRate = getCurrentConfigData().skidYawMax
	skidSteerBoost = getCurrentConfigData().skidYawBoost
	
	
	dump(yaw_PID)
	
	dump(getCurrentConfigData())
	
	
end



local function serialize()
  return {escConfigKey = currentESCConfigurationKey}
end


local function deserialize(data)
  if data and data.escConfigKey then
    setESCMode(data.escConfigKey)
  end
end


-- public interface
M.init = init
M.initSecondStage = initSecondStage
M.updateWheelsIntermediate = nil
M.updateGFX = nil
M.toggleESCMode = toggleESCMode
M.nextESCMode = nextESCMode
M.previousESCMode = previousESCMode
M.setESCMode = setESCMode
M.getCarData = getCarData
M.getCurrentConfigData = getCurrentConfigData
M.startESCCalibration = startESCCalibration
M.stopESCCalibration = stopESCCalibration
M.serialize = serialize
M.deserialize = deserialize

return M